@@ -10,6 +10,23 @@ const GLib = imports.gi.GLib;
 const Gio = imports.gi.Gio;
 const GObject = imports.gi.GObject;
 
+function expectWarn64(callable) {
+    GLib.test_expect_message('Gjs', GLib.LogLevelFlags.LEVEL_WARNING,
+        '*cannot be safely stored*');
+    const ret = callable();
+    GLib.test_assert_expected_messages_internal('Gjs',
+        'testRegress.js', 0, 'Ignore message');
+    return ret;
+}
+
+const bit64Types = ['uint64', 'int64'];
+if (GLib.SIZEOF_LONG === 8)
+    bit64Types.push('long', 'ulong');
+if (GLib.SIZEOF_SIZE_T === 8)
+    bit64Types.push('size');
+if (GLib.SIZEOF_SSIZE_T === 8)
+    bit64Types.push('ssize');
+
 describe('Life, the Universe and Everything', function () {
     it('includes null return value', function () {
         expect(Regress.test_return_allow_none()).toBeNull();
@@ -29,12 +46,28 @@ describe('Life, the Universe and Everything', function () {
             expect(Regress[method](42)).toBe(42);
             expect(Regress[method](-42)).toBe(-42);
             expect(Regress[method](undefined)).toBe(0);
+            expect(Regress[method](42.42)).toBe(42);
+            expect(Regress[method](-42.42)).toBe(-42);
+
+            if (bits >= 64) {
+                expect(Regress[method](42n)).toBe(42);
+                expect(Regress[method](-42n)).toBe(-42);
+            } else {
+                expect(() => Regress[method](42n)).toThrow();
+                expect(() => Regress[method](-42n)).toThrow();
+            }
         });
 
         it(`includes unsigned ${bits}-bit integers`, function () {
             const method = `test_uint${bits}`;
             expect(Regress[method](42)).toBe(42);
             expect(Regress[method](undefined)).toBe(0);
+            expect(Regress[method](42.42)).toBe(42);
+
+            if (bits >= 64)
+                expect(Regress[method](42n)).toEqual(42);
+            else
+                expect(() => Regress[method](42n)).toThrow();
         });
     });
 
@@ -44,10 +77,23 @@ describe('Life, the Universe and Everything', function () {
             expect(Regress[method](42)).toBe(42);
             expect(Regress[method](-42)).toBe(-42);
 
-            if (['float', 'double'].includes(type))
+            if (['float', 'double'].includes(type)) {
                 expect(Number.isNaN(Regress[method](undefined))).toBeTruthy();
-            else
+                expect(Regress[method](42.42)).toBeCloseTo(42.42);
+                expect(Regress[method](-42.42)).toBeCloseTo(-42.42);
+            } else {
                 expect(Regress[method](undefined)).toBe(0);
+                expect(Regress[method](42.42)).toBe(42);
+                expect(Regress[method](-42.42)).toBe(-42);
+            }
+
+            if (bit64Types.includes(type)) {
+                expect(Regress[method](42n)).toBe(42);
+                expect(Regress[method](-42n)).toBe(-42);
+            } else {
+                expect(() => Regress[method](42n)).toThrow();
+                expect(() => Regress[method](-42n)).toThrow();
+            }
         });
     });
 
@@ -56,17 +102,81 @@ describe('Life, the Universe and Everything', function () {
             const method = `test_${type}`;
             expect(Regress[method](42)).toBe(42);
             expect(Regress[method](undefined)).toBe(0);
+            expect(Regress[method](42.42)).toBe(42);
+
+            if (bit64Types.includes(type))
+                expect(Regress[method](42n)).toBe(42);
+            else
+                expect(() => Regress[method](42n)).toThrow();
         });
     });
 
     describe('No implicit conversion to unsigned', function () {
         ['uint8', 'uint16', 'uint32', 'uint64', 'uint', 'size'].forEach(type => {
             it(`for ${type}`, function () {
-                expect(() => Regress[`test_${type}`](-42)).toThrow();
+                expect(() => Regress[`test_${type}`](-42)).toThrowError(/out of range/);
+
+                if (bit64Types.includes(type))
+                    expect(() => Regress[`test_${type}`](-42n)).toThrowError(/out of range/);
+                else
+                    expect(() => Regress[`test_${type}`](-42n)).toThrow();
             });
         });
     });
 
+    describe('Infinity and NaN', function () {
+        ['int8', 'int16', 'int32', 'int64', 'short', 'int', 'long', 'ssize'].forEach(type => {
+            it(`converts to 0 for ${type}`, function () {
+                expect(Regress[`test_${type}`](Infinity)).toBe(0);
+                expect(Regress[`test_${type}`](-Infinity)).toBe(0);
+                expect(Regress[`test_${type}`](NaN)).toBe(0);
+            });
+        });
+
+        ['uint8', 'uint16', 'uint32', 'uint64', 'ushort', 'uint', 'ulong', 'size'].forEach(type => {
+            it(`converts to 0 for ${type}`, function () {
+                expect(Regress[`test_${type}`](Infinity)).toBe(0);
+                expect(Regress[`test_${type}`](NaN)).toBe(0);
+            });
+        });
+
+        ['float', 'double'].forEach(type => {
+            it(`not for ${type}`, function () {
+                expect(Regress[`test_${type}`](Infinity)).toBe(Infinity);
+                expect(Regress[`test_${type}`](-Infinity)).toBe(-Infinity);
+                expect(Regress[`test_${type}`](NaN)).toBeNaN();
+            });
+        });
+    });
+
+    describe('(u)int64 numeric values', function () {
+        const minInt64 = -(2n ** 63n);
+        const maxInt64 = 2n ** 63n - 1n;
+        const maxUint64 = 2n ** 64n - 1n;
+
+        ['uint64', 'int64', 'long', 'ulong', 'size', 'ssize'].forEach(type => {
+            if (!bit64Types.includes(type))
+                return;
+            const signed = ['int64', 'long', 'ssize'].includes(type);
+            const limits = {
+                min: signed ? minInt64 : 0n,
+                max: signed ? maxInt64 : maxUint64,
+            };
+            const testFunc = Regress[`test_${type}`];
+
+            it(`can use numeric limits for ${type}`, function () {
+                expect(expectWarn64(() => testFunc(limits.max)))
+                    .toEqual(Number(limits.max));
+
+                if (signed) {
+                    expect(expectWarn64(() => testFunc(limits.min)))
+                        .toEqual(Number(limits.min));
+                }
+            });
+        });
+    });
+
+
     it('includes wide characters', function () {
         expect(Regress.test_unichar('c')).toBe('c');
         expect(Regress.test_unichar('')).toBe('');
@@ -285,6 +395,11 @@ describe('Life, the Universe and Everything', function () {
         });
     });
 
+    it('integer array with static length', function () {
+        const arr = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10];
+        expect(() => Regress.test_array_static_in_int(arr)).not.toThrow();
+    });
+
     it("string array that's const in C", function () {
         expect(Regress.test_strv_out_c()).toEqual(['thanks', 'for', 'all', 'the', 'fish']);
     });
@@ -925,6 +1040,113 @@ describe('Life, the Universe and Everything', function () {
             expect(v).toEqual(jasmine.any(Regress.TestObj));
         });
 
+        describe('GProperty', function () {
+            let t, boxed, hashTable, hashTable2, list2, string, gtype, byteArray;
+            const list = null;
+            const int = 42;
+            const double = Math.PI;
+            const double2 = Math.E;
+
+            beforeEach(function () {
+                boxed = new Regress.TestBoxed({some_int8: 127});
+                hashTable = {a: 1, b: 2};
+                hashTable2 = {c: 3, d: 4};
+                list2 = ['j', 'k', 'l'];
+                string = 'cauliflower';
+                gtype = GObject.Object.$gtype;
+                byteArray = Uint8Array.from('abcd', c => c.charCodeAt(0));
+                t = new Regress.TestObj({
+                    boxed,
+                    // hashTable,
+                    list,
+                    // pptrarray: list,
+                    // hashTableOld: hashTable,
+                    listOld: list,
+                    int,
+                    float: double,
+                    double,
+                    string,
+                    gtype,
+                    // byteArray,
+                });
+            });
+
+            it('Boxed type', function () {
+                expect(t.boxed.some_int8).toBe(127);
+                const boxed2 = new Regress.TestBoxed({some_int8: 31});
+                t.boxed = boxed2;
+                expect(t.boxed.some_int8).toBe(31);
+            });
+
+            xit('Hash table', function () {
+                expect(t.hashTable).toBe(hashTable);
+                t.hashTable = hashTable2;
+                expect(t.hashTable).toBe(hashTable2);
+            }).pend('https://gitlab.gnome.org/GNOME/gjs/-/issues/83');
+
+            xit('List', function () {
+                expect(t.list).toBe(list);
+                t.list = list2;
+                expect(t.list).toBe(list2);
+            }).pend('https://gitlab.gnome.org/GNOME/gjs/-/issues/83');
+
+            xit('Pointer array', function () {
+                expect(t.pptrarray).toBe(list);
+                t.pptrarray = list2;
+                expect(t.pptrarray).toBe(list2);
+            }).pend('https://gitlab.gnome.org/GNOME/gjs/-/issues/83');
+
+            xit('Hash table with old-style annotation', function () {
+                expect(t.hashTableOld).toBe(hashTable);
+                t.hashTableOld = hashTable2;
+                expect(t.hashTableOld).toBe(hashTable2);
+            }).pend('https://gitlab.gnome.org/GNOME/gjs/-/issues/83');
+
+            xit('List with old-style annotation', function () {
+                expect(t.listOld).toBe(list);
+                t.listOld = list2;
+                expect(t.listOld).toBe(list2);
+            }).pend('https://gitlab.gnome.org/GNOME/gjs/-/issues/83');
+
+            it('Integer', function () {
+                expect(t.int).toBe(int);
+                t.int = 35;
+                expect(t.int).toBe(35);
+            });
+
+            it('Float', function () {
+                expect(t.float).toBeCloseTo(double);
+                t.float = double2;
+                expect(t.float).toBeCloseTo(double2);
+            });
+
+            it('Double', function () {
+                expect(t.double).toBeCloseTo(double);
+                t.double = double2;
+                expect(t.double).toBeCloseTo(double2);
+            });
+
+            it('String', function () {
+                expect(t.string).toBe(string);
+                t.string = 'string2';
+                expect(t.string).toBe('string2');
+            });
+
+            xit('GType object', function () {
+                expect(t.gtype).toBe(gtype);
+                const gtype2 = GObject.InitiallyUnowned.$gtype;
+                t.gtype = gtype2;
+                expect(t.gtype).toBe(gtype2);
+            }).pend('https://gitlab.gnome.org/GNOME/gjs/-/issues/83');
+
+            xit('Byte array', function () {
+                expect(t.byteArray).toBe(byteArray);
+                const byteArray2 = Uint8Array.from('efgh', c => c.charCodeAt(0));
+                t.byteArray = byteArray2;
+                expect(t.byteArray).toBe(byteArray2);
+            }).pend('https://gitlab.gnome.org/GNOME/gjs/-/issues/276');
+        });
+
         describe('Object-valued GProperty', function () {
             let o1, t1, t2;
             beforeEach(function () {
@@ -1250,6 +1472,7 @@ describe('Life, the Universe and Everything', function () {
                 int: 42,
                 float: Math.PI,
                 double: Math.E,
+                boolean: true,
             });
         });
 
@@ -1267,6 +1490,12 @@ describe('Life, the Universe and Everything', function () {
         it('can call an instance method that overrides the parent class', function () {
             expect(subobj.instance_method()).toEqual(0);
         });
+
+        it('can have its own properties', function () {
+            expect(subobj.boolean).toBeTruthy();
+            subobj.boolean = false;
+            expect(subobj.boolean).toBeFalsy();
+        });
     });
 
     describe('Overridden properties on interfaces', function () {
@@ -1411,6 +1640,12 @@ describe('Life, the Universe and Everything', function () {
         expect(callback).toHaveBeenCalled();
     });
 
+    it('static method taking a callback', function () {
+        const callback = jasmine.createSpy('callback');
+        Regress.TestObj.static_method_callback(callback);
+        expect(callback).toHaveBeenCalled();
+    });
+
     it('constructor taking a callback', function () {
         const callback = jasmine.createSpy('callback').and.returnValue(42);
         void Regress.TestObj.new_callback(callback);
