@@ -30,7 +30,7 @@ do_Get_Upstream_Base () {
     # of date branches.
     #
     # Limit the fetch to a certain date horizon to limit the amount of data we
-    # get. If the branch was forked from origin/master before this horizon, it
+    # get. If the branch was forked from the main branch before this horizon, it
     # should probably be rebased.
     git remote add upstream https://gitlab.gnome.org/GNOME/gjs.git || \
         git remote set-url upstream https://gitlab.gnome.org/GNOME/gjs.git
@@ -116,7 +116,7 @@ fi
 # Show some environment info
 do_Print_Labels  'ENVIRONMENT'
 echo "Running on: $BASE $OS"
-echo "Job: $TASK_ID"
+echo "Job: $CI_JOB_NAME"
 echo "Configure options: ${CONFIG_OPTS-$BUILD_OPTS}"
 echo "Doing: $1 $extra_opts"
 
